/*
 * Probability distributions.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: distributions.cc,v 2.1 2005/02/08 09:20:34 lorens Exp $
 */
#include "distributions.h"
#include "rational.h"
#include "rng.h"
#include <stdexcept>


/*
 * Inverse error function.
 */
static double erfinv(double y) {
  // MATLAB code
  static double a[] = { 0.886226899, -1.645349621, 0.914624893, -0.140543331 };
  static double b[] = { -2.118377725, 1.442710462, -0.329097515, 0.012229801 };
  static double c[] = { -1.970840454, -1.624906493, 3.429567803, 1.641345311 };
  static double d[] = { 3.543889200, 1.637067800 };

  if (y == -1.0) {
    return -HUGE_VAL;
  } else if (y == 1.0) {
    return HUGE_VAL;
  } else if (isnan(y) || fabs(y) > 1.0) {
    return NAN;
  } else {
    double x;
    if (fabs(y) <= 0.7) {
      double z = y*y;
      x = y*(((a[3]*z + a[2])*z + a[1])*z + a[0]);
      x /= (((b[3]*z + b[2])*z + b[1])*z + b[0])*z + 1.0;
    } else if (y > 0.7 && y < 1.0) {
      double z = sqrt(-log(0.5*(1.0 - y)));
      x = (((c[3]*z + c[2])*z + c[1])*z + c[0])/((d[1]*z + d[0])*z + 1.0);
    } else { /* y < -0.7 && y > -1.0 */
      double z = sqrt(-log(0.5*(1.0 + y)));
      x = -(((c[3]*z + c[2])*z + c[1])*z + c[0])/((d[1]*z + d[0])*z + 1.0);
    }
    double u = (erf(x) - y)/(M_2_SQRTPI*exp(-x*x));
    return x - u/(1.0 + x*u);
  }
}


/* ====================================================================== */
/* Distribution */

/* The standard exponential distribution: Exp(1). */
const Distribution& Distribution::EXP1 = Exponential::EXP1_;


/* Conditional probability density function for this distribution. */
double Distribution::pdf(double t, double t0) const {
  double p0 = cdf(t0);
  if (p0 >= 1.0) {
    throw std::out_of_range("cdf(t0) is 1 in pdf(t, t0)");
  } else if (t <= 0.0) {
    return 0.0;
  } else {
    return pdf(t + t0)/(1.0 - p0);
  }
}


/* Conditional cumulative distribution function for this distribution. */
double Distribution::cdf(double t, double t0) const {
  double p0 = cdf(t0);
  if (p0 >= 1.0) {
    throw std::out_of_range("cdf(t0) is 1 in cdf(t, t0)");
  } else if (t <= 0) {
    return 0.0;
  } else {
    return 1.0 - (1.0 - cdf(t + t0))/(1.0 - p0);
  }
}


/* Conditional inverse cumulative distribution function for this
   distribution. */
double Distribution::inv(double p, double t0) const {
  double p0 = cdf(t0);
  if (p0 >= 1.0) {
    throw std::out_of_range("cdf(t0) is 1 in inv(t, t0)");
  } else if (p < 0.0 || p > 1.0) {
    throw std::out_of_range("p < 0 or p > 1 in inv(p, t0)");
  } else {
    return inv(1.0 - (1.0 - p)*(1.0 - p0)) - t0;
  }
}


/* Returns a sample drawn from this distribution. */
double Distribution::sample(double t0) const {
  return inv(genrand_real3(), t0) + t0;
}


/* Output operator for distributions. */
std::ostream& operator<<(std::ostream& os, const Distribution& d) {
  d.print(os);
  return os;
}


/* ====================================================================== */
/* Exponential */

/* The standard exponential distribution: Exp(1). */
const Exponential Exponential::EXP1_;


/* Returns an exponential distribution with the given rate. */
const Exponential& Exponential::make(const Rational& rate) {
  if (rate == 1) {
    return EXP1_;
  } else {
    return *new Exponential(rate.double_value());
  }
}


/* Constructs an exponential distribution with rate 1. */
Exponential::Exponential()
  : rate_(1.0) {
  ref(this);
}


/* Probability density function for this distribution. */
double Exponential::pdf(double t) const {
  if (t < 0.0) {
    return 0.0;
  } else {
    return rate()*exp(-rate()*t);
  }
}


/* Cumulative distribution function for this distribution. */
double Exponential::cdf(double t) const {
  if (t <= 0.0) {
    return 0.0;
  } else {
    return 1.0 - exp(-rate()*t);
  }
}


/* Inverse cumulative distribution function for this distribution. */
double Exponential::inv(double p) const {
  if (p <= 0.0) {
    return 0.0;
  } else if (p >= 1.0) {
    return HUGE_VAL;
  } else {
    return -log(1.0 - p)/rate();
  }
}


/* Fills the provided list with the first n moments of this distribution. */
void Exponential::moments(std::vector<double>& m, size_t n) const {
  double lambda_inv = 1.0/rate();
  double mi = 1.0;
  for (size_t i = 1; i <= n; i++) {
    mi *= i*lambda_inv;
    m.push_back(mi);
  }
}


/* Prints this object on the given stream. */
void Exponential::print(std::ostream& os) const {
  os << "(exponential " << rate() << ")";
}


/* ===================================================================== */
/* Weibull */

/* Returns a Weibull distribution with the given scale and shape. */
const Distribution& Weibull::make(const Rational& scale,
				  const Rational& shape) {
  if (shape == 1) {
    return Exponential::make(1/scale);
  } else {
    return *new Weibull(scale.double_value(), shape.double_value());
  }
}


/* Probability density function for this distribution. */
double Weibull::pdf(double t) const {
  if (t < 0.0) {
    return 0.0;
  } else {
    double x = pow(t/scale(), shape());
    return shape()*x/t*exp(-x);
  }
}


/* Cumulative distribution function for this distribution. */
double Weibull::cdf(double t) const {
  if (t <= 0.0) {
    return 0.0;
  } else {
    return 1.0 - exp(-pow(t/scale(), shape()));
  }
}


/* Inverse cumulative distribution function for this distribution. */
double Weibull::inv(double p) const {
  if (p <= 0.0) {
    return 0.0;
  } else if (p >= 1.0) {
    return HUGE_VAL;
  } else {
    return scale()*pow(-log(1.0 - p), 1.0/shape());
  }
}


/* Fills the provided list with the first n moments of this distribution. */
void Weibull::moments(std::vector<double>& m, size_t n) const {
  /* N.B. this function should never be called for a distribution with
     non-constant parameters. */
  double eta = scale();
  double beta_inv = 1.0/shape();
  double ei = 1.0;
  double bi = 1.0;
  for (size_t i = 1; i <= n; i++) {
    ei *= eta;
    bi += beta_inv;
    m.push_back(ei*tgamma(bi));
  }
}


/* Prints this object on the given stream. */
void Weibull::print(std::ostream& os) const {
  os << "(weibull " << scale() << ' ' << shape() << ")";
}


/* ===================================================================== */
/* Lognormal */

/* Returns a lognormal distribution with the given scale and shape. */
const Lognormal& Lognormal::make(const Rational& scale,
				 const Rational& shape) {
  return *new Lognormal(scale.double_value(), shape.double_value());
}


/* Probability density function for this distribution. */
double Lognormal::pdf(double t) const {
  if (t < 0.0) {
    return 0.0;
  } else {
    double x = log(t/scale()) + 0.5*shape()*shape();
    return 1.0/shape()/sqrt(2.0*M_PI)/t*exp(-0.5*x*x/shape()/shape());
  }
}


/* Cumulative distribution function for this distribution. */
double Lognormal::cdf(double t) const {
  if (t <= 0.0) {
    return 0.0;
  } else {
    return 0.5*(1.0 + erf((log(t/scale())/shape() + 0.5*shape())*M_SQRT1_2));
  }
}


/* Inverse cumulative distribution function for this distribution. */
double Lognormal::inv(double p) const {
  if (p <= 0.0) {
    return 0.0;
  } else if (p >= 1.0) {
    return HUGE_VAL;
  } else {
    return scale()*exp(shape()*(M_SQRT2*erfinv(2.0*p - 1.0) - 0.5*shape()));
  }
}


/* Fills the provided list with the first n moments of this distribution. */
void Lognormal::moments(std::vector<double>& m, size_t n) const {
  /* N.B. this function should never be called for a distribution with
     non-constant parameters. */
  double mu = scale();
  double sigma = shape();
  double mean = log(mu) - sigma*sigma/2.0;
  for (size_t i = 1; i <= n; i++) {
    m.push_back(exp(i*mean + i*i*sigma*sigma/2.0));
  }
}


/* Prints this object on the given stream. */
void Lognormal::print(std::ostream& os) const {
  os << "(lognormal " << scale() << ' ' << shape() << ")";
}


/* ===================================================================== */
/* Uniform */

/* Returns a uniform distribution with the given bounds. */
const Uniform& Uniform::make(const Rational& low, const Rational& high) {
  return *new Uniform(low.double_value(), high.double_value());
}


/* Fills the provided list with the first n moments of this distribution. */
void Uniform::moments(std::vector<double>& m, size_t n) const {
  /* N.B. this function should never be called for a distribution with
     non-constant parameters. */
  double a = low();
  double b = high();
  double ai = a;
  double bi = b;
  for (size_t i = 1; i <= n; i++) {
    ai *= a;
    bi *= b;
    m.push_back((bi - ai)/((i + 1)*(b - a)));
  }
}


/* Probability density function for this distribution. */
double Uniform::pdf(double t) const {
  if (t < low() || t > high()) {
    return 0.0;
  } else {
    return 1.0/(high() - low());
  }
}


/* Cumulative distribution function for this distribution. */
double Uniform::cdf(double t) const {
  if (t <= low()) {
    return 0.0;
  } else if (t >= high()) {
    return 1.0;
  } else {
    return (t - low())/(high() - low());
  }
}


/* Inverse cumulative distribution function for this distribution. */
double Uniform::inv(double p) const {
  if (p <= 0.0) {
    return low();
  } else if (p >= 1.0) {
    return high();
  } else {
    return (high() - low())*p + low();
  }
}


/* Prints this object on the given stream. */
void Uniform::print(std::ostream& os) const {
  os << "(uniform " << low() << ' ' << high() << ")";
}
