/* -*-C++-*- */
/*
 * Probability distributions.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: distributions.h,v 2.1 2005/02/08 09:22:41 lorens Exp $
 */
#ifndef DISTRIBUTIONS_H
#define DISTRIBUTIONS_H

#include <config.h>
#include "refcount.h"
#include <iostream>
#include <map>
#include <utility>
#include <vector>

struct Rational;


/* ====================================================================== */
/* Distribution */

/*
 * A probability distribution.
 */
struct Distribution : public RCObject {
  /* The standard exponential distribution: Exp(1). */
  static const Distribution& EXP1;

  /* Probability density function for this distribution. */
  virtual double pdf(double t) const = 0;

  /* Conditional probability density function for this distribution. */
  double pdf(double t, double t0) const;

  /* Cumulative distribution function for this distribution. */
  virtual double cdf(double t) const = 0;

  /* Conditional cumulative distribution function for this distribution. */
  double cdf(double t, double t0) const;

  /* Inverse cumulative distribution function for this distribution. */
  virtual double inv(double p) const = 0;

  /* Conditional inverse cumulative distribution function for this
     distribution. */
  double inv(double p, double t0) const;

  /* Fills the provided list with the first n moments of this distribution. */
  virtual void moments(std::vector<double>& m, size_t n) const = 0;

  /* Returns a sample drawn from this distribution. */
  double sample(double t0 = 0.0) const;

protected:
  /* Constructs a distribution. */
  Distribution() {}

  /* Prints this object on the given stream. */
  virtual void print(std::ostream& os) const = 0;

  friend std::ostream& operator<<(std::ostream& os, const Distribution& d);
};

/* Output operator for distributions. */
std::ostream& operator<<(std::ostream& os, const Distribution& d);


/* ====================================================================== */
/* Exponential */

/*
 * An exponential distribution.
 */
struct Exponential : public Distribution {
  /* Returns an exponential distribution with the given rate. */
  static const Exponential& make(const Rational& rate);

  /* Constructs an exponential distribution with the given rate. */
  explicit Exponential(double rate) : rate_(rate) {}

  /* Returns the rate of this exponential distribution. */
  double rate() const { return rate_; }

  /* Probability density function for this distribution. */
  virtual double pdf(double t) const;

  /* Cumulative distribution function for this distribution. */
  virtual double cdf(double t) const;

  /* Inverse cumulative distribution function for this distribution. */
  virtual double inv(double p) const;

  /* Fills the provided list with the first n moments of this distribution. */
  virtual void moments(std::vector<double>& m, size_t n) const;

protected:
  /* Prints this object on the given stream. */
  virtual void print(std::ostream& os) const;

private:
  /* The standard exponential distribution: Exp(1). */
  static const Exponential EXP1_;

  /* The rate of this exponential distribution. */
  double rate_;

  /* Constructs an exponential distribution with rate 1. */
  Exponential();

  friend struct Distribution;
};


/* ====================================================================== */
/* Weibull */

/*
 * A Weibull distribution.
 */
struct Weibull : public Distribution {
  /* Returns a Weibull distribution with the given scale and shape. */
  static const Distribution& make(const Rational& scale,
				  const Rational& shape);

  /* Constructs a Weibull distribution with the given scale and shape. */
  Weibull(double scale, double shape) : scale_(scale), shape_(shape) {}

  /* Returns the scale of this Weibull distribution. */
  double scale() const { return scale_; }

  /* Returns the shape of this Weibull distribution. */
  double shape() const { return shape_; }

  /* Probability density function for this distribution. */
  virtual double pdf(double t) const;

  /* Cumulative distribution function for this distribution. */
  virtual double cdf(double t) const;

  /* Inverse cumulative distribution function for this distribution. */
  virtual double inv(double p) const;

  /* Fills the provided list with the first n moments of this distribution. */
  virtual void moments(std::vector<double>& m, size_t n) const;

protected:
  /* Prints this object on the given stream. */
  virtual void print(std::ostream& os) const;

private:
  /* The scale of this Weibull distribution. */
  double scale_;
  /* The shape of this Weibull distribution. */
  double shape_;
};


/* ====================================================================== */
/* Lognormal */

/*
 * A lognormal distribution.
 */
struct Lognormal : public Distribution {
  /* Returns a lognormal distribution with the given scale and shape. */
  static const Lognormal& make(const Rational& scale,
			       const Rational& shape);

  /* Constructs a lognormal distribution with the given scale and shape. */
  Lognormal(double scale, double shape)
    : scale_(scale), shape_(shape) {}

  /* Returns the scale of this lognormal distribution. */
  double scale() const { return scale_; }

  /* Returns the shape of this lognormal distribution. */
  double shape() const { return shape_; }

  /* Probability density function for this distribution. */
  virtual double pdf(double t) const;

  /* Cumulative distribution function for this distribution. */
  virtual double cdf(double t) const;

  /* Inverse cumulative distribution function for this distribution. */
  virtual double inv(double p) const;

  /* Fills the provided list with the first n moments of this distribution. */
  virtual void moments(std::vector<double>& m, size_t n) const;

protected:
  /* Prints this object on the given stream. */
  virtual void print(std::ostream& os) const;

private:
  /* The scale of this lognormal distribution. */
  double scale_;
  /* The shape of this lognormal distribution. */
  double shape_;
};


/* ====================================================================== */
/* Uniform */

/*
 * A uniform distribution.
 */
struct Uniform : public Distribution {
  /* Returns a uniform distribution with the bounds. */
  static const Uniform& make(const Rational& low, const Rational& high);

  /* Constructs a uniform distribution with the given bounds. */
  Uniform(double low, double high) : low_(low), high_(high) {}

  /* Returns the lower bound of this uniform distribution. */
  double low() const { return low_; }

  /* Returns the upper bound of this uniform distribution. */
  double high() const { return high_; }

  /* Probability density function for this distribution. */
  virtual double pdf(double t) const;

  /* Cumulative distribution function for this distribution. */
  virtual double cdf(double t) const;

  /* Inverse cumulative distribution function for this distribution. */
  virtual double inv(double p) const;

  /* Fills the provided list with the first n moments of this distribution. */
  virtual void moments(std::vector<double>& m, size_t n) const;

protected:
  /* Prints this object on the given stream. */
  virtual void print(std::ostream& os) const;

private:
  /* The lower bound of this uniform distribution. */
  double low_;
  /* The upper bound of this uniform distribution. */
  double high_;
};


#endif /* DISTRIBUTIONS_H */
