/*
 * Domain descriptions.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: domains.cc,v 2.2 2005/02/08 09:25:02 lorens Exp $
 */
#include "domains.h"


/* ====================================================================== */
/* Domain */

/* Table of defined domains. */
Domain::DomainMap Domain::domains = Domain::DomainMap();


/* Returns a const_iterator pointing to the first domain. */
Domain::DomainMap::const_iterator Domain::begin() {
  return domains.begin();
}


/* Returns a const_iterator pointing beyond the last domain. */
Domain::DomainMap::const_iterator Domain::end() {
  return domains.end();
}


/* Returns a pointer to the domain with the given name, or 0 if it
   is undefined. */
const Domain* Domain::find(const std::string& name) {
  DomainMap::const_iterator di = domains.find(name);
  return (di != domains.end()) ? (*di).second : 0;
}


/* Removes all defined domains. */
void Domain::clear() {
  DomainMap::const_iterator di = begin();
  while (di != end()) {
    delete (*di).second;
    di = begin();
  }
  domains.clear();
}


/* Constructs an empty domain with the given name. */
Domain::Domain(const std::string& name)
  : name_(name), total_time_(functions_.add_function("total-time")),
    goal_achieved_(functions_.add_function("goal-achieved")) {
  const Domain* d = find(name);
  if (d != 0) {
    delete d;
  }
  domains[name] = this;
  FunctionTable::make_dynamic(total_time_);
  FunctionTable::make_dynamic(goal_achieved_);
}


/* Deletes a domain. */
Domain::~Domain() {
  domains.erase(name());
  for (EventSchemaMap::const_iterator ei = events().begin();
       ei != events().end(); ei++) {
    delete (*ei).second;
  }
}


/* Adds the given event to this domain. */
void Domain::add_event(const EventSchema& event) {
  events_.insert(std::make_pair(event.name(), &event));
}


/* Returns the event with the given name, or 0 if there is no event
   with the given name. */
const EventSchema* Domain::find_event(const std::string& name) const {
  EventSchemaMap::const_iterator ei = events_.find(name);
  return (ei != events_.end()) ? (*ei).second : 0;
}


/* Output operator for domains. */
std::ostream& operator<<(std::ostream& os, const Domain& d) {
  os << "name: " << d.name();
  os << std::endl << "types:" << d.types();
  os << std::endl << "constants:" << d.terms();
  os << std::endl << "predicates:" << d.predicates();
  os << std::endl << "functions:" << d.functions();
  os << std::endl << "events:";
  for (EventSchemaMap::const_iterator ei = d.events().begin();
       ei != d.events().end(); ei++) {
    os << std::endl << *(*ei).second;
  }
  return os;
}
