/*   EMpht.c  */
/*Latest corrections made 9 March, 1998  */
/* Adapted for Tempastic-DTP. */
#include "phfit.h"
#include "rng.h"
#include <unistd.h>
#include <iostream>
#include <vector>

#define PMIN 1e-6


struct dvec_t : public std::vector<double> {
  dvec_t() : std::vector<double>() {}
  dvec_t(size_t n, double d = 0.0) : std::vector<double>(n, d) {}
};


struct dmat_t : public std::vector<std::vector<double> > {
  dmat_t(size_t m, size_t n, double d = 0.0)
    : std::vector<std::vector<double> >(m, std::vector<double>(n, d)) {}
};


/* Verbosity level. */
extern int verbosity;

static dvec_t obs, weight;
double weight_sum;


static void show_pi_T(const dvec_t& pi, const dmat_t& T, const dvec_t& t) {
  size_t p = pi.size();
  for (size_t i = 0; i < p; i++) {
    std::cout << std::endl << pi[i] << ' ';
    for (size_t j = 0; j < p; j++) {
      std::cout << ' ' << T[i][j];
    }
    std::cout << "  " << t[i];
  }
  std::cout << std::endl;
}


static void a_rungekutta(dvec_t& avector,
			 double dt, double h, const dmat_t& T) {
  size_t p = avector.size();
  dvec_t ka0(p), ka1(p), ka2(p), ka3(p);
  double h2 = dt/(trunc(dt/h) + 1.0);
  for (double eps = 0.0; eps <= dt - 0.5*h2; eps += h2) {
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[j][i]*avector[j];
      }
      ka0[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[j][i]*(avector[j] + 0.5*ka0[j]);
      }
      ka1[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[j][i]*(avector[j] + 0.5*ka1[j]);
      }
      ka2[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[j][i]*(avector[j] + ka2[j]);
      }
      ka3[i] = h2*sum;
    }

    for (size_t i = 0; i < p; i++) {
      avector[i] += (ka0[i] + 2.0*ka1[i] + 2.0*ka2[i] + ka3[i])/6.0;
    }
  }
}


static void rungekutta(dvec_t& avector, dvec_t& bvector, dmat_t& cmatrix,
		       double dt, double h, const dmat_t& T, const dvec_t& t) {
  size_t p = avector.size();
  dvec_t ka0(p), ka1(p), ka2(p), ka3(p);
  dvec_t kb0(p), kb1(p), kb2(p), kb3(p);
  dmat_t kc0(p, p), kc1(p, p), kc2(p, p), kc3(p, p);
  double h2 = dt/(trunc(dt/h) + 1.0);
  for (double eps = 0.0; eps <= dt - 0.5*h2; eps += h2) {
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[j][i]*avector[j];
      }
      ka0[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[j][i]*(avector[j] + 0.5*ka0[j]);
      }
      ka1[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[j][i]*(avector[j] + 0.5*ka1[j]);
      }
      ka2[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[j][i]*(avector[j] + ka2[j]);
      }
      ka3[i] = h2*sum;
    }
    
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[i][j]*bvector[j];
      }
      kb0[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[i][j]*(bvector[j] + 0.5*kb0[j]);
      }
      kb1[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j=0; j < p; j++) {
        sum += T[i][j]*(bvector[j] + 0.5*kb1[j]);
      }
      kb2[i] = h2*sum;
    }
    for (size_t i = 0; i < p; i++) {
      double sum = 0.0;
      for (size_t j = 0; j < p; j++) {
        sum += T[i][j]*(bvector[j] + kb2[j]);
      }
      kb3[i] = h2*sum;
    }
   
    for (size_t m = 0; m < p; m++) {
      for (size_t i = 0; i < p; i++) {
        double sum = t[m]*avector[i];
        for (size_t j = 0; j < p; j++) {
          sum += T[m][j]*cmatrix[j][i];
	}
        kc0[m][i] = h2*sum;
      }
    }
    for (size_t m = 0; m < p; m++) {
      for (size_t i = 0; i < p; i++) {
        double sum = t[m]*(avector[i] + 0.5*ka0[i]);
        for (size_t j = 0; j < p; j++) {
          sum += T[m][j]*(cmatrix[j][i] + 0.5*kc0[j][i]);
	}
        kc1[m][i] = h2*sum;
      }
    }
    for (size_t m = 0; m < p; m++) {
      for (size_t i = 0; i < p; i++) {
        double sum = t[m]*(avector[i] + 0.5*ka1[i]);
        for (size_t j = 0; j < p; j++) {
          sum += T[m][j]*(cmatrix[j][i] + 0.5*kc1[j][i]);
	}
        kc2[m][i] = h2*sum;
      }
    }
    for (size_t m = 0; m < p; m++) {
      for (size_t i=0; i < p; i++) {
        double sum = t[m]*(avector[i] + ka2[i]);
        for (size_t j=0; j < p; j++) {
          sum += T[m][j]*(cmatrix[j][i] + kc2[j][i]);
	}
        kc3[m][i] = h2*sum;
      }
    }

    for (size_t i = 0; i < p; i++) {
      avector[i] += (ka0[i] + 2.0*ka1[i] + 2.0*ka2[i] + ka3[i])/6.0;
      bvector[i] += (kb0[i] + 2.0*kb1[i] + 2.0*kb2[i] + kb3[i])/6.0;
      for (size_t j = 0; j < p; j++) {
        cmatrix[i][j] +=(kc0[i][j] + 2.0*kc1[i][j]
			 + 2.0*kc2[i][j] + kc3[i][j])/6.0;
      }
    }
  }
}


static double set_step_length(const dmat_t& T) {    
  size_t p = T.size();
  double h = -0.1/T[0][0];
  for (size_t i = 1; i < p; i++) {
    if (h > -0.1/T[i][i]) {
      h = -0.1/T[i][i];
    }
  }
  return h;
}


static void EMstep(dvec_t& pi, dmat_t& T, dvec_t& t, double h) {
  size_t p = pi.size();
  dvec_t Bmean(p);
  dvec_t Zmean(p);
  dmat_t Nmean(p, p+1);
  if (obs.size() > 0) {
    dvec_t avector(pi), bvector(t);
    dmat_t cmatrix(p, p);
    double dt = obs[0];
    for (size_t k = 0; k < obs.size(); k++) {
      if (dt > 0) {
	rungekutta(avector, bvector, cmatrix, dt, h, T, t);
      }
      double pitimesb = 0.0;
      for (size_t i = 0; i < p; i++) {
	pitimesb += pi[i]*bvector[i];
      }
      for (size_t i = 0; i < p; i++) {
	Bmean[i] += pi[i]*bvector[i]*weight[k]/pitimesb;
	Nmean[i][p] += avector[i]*t[i]*weight[k]/pitimesb;
	Zmean[i] += cmatrix[i][i]*weight[k]/pitimesb;
	for (size_t j = 0; j < p; j++) {
	  Nmean[i][j] += T[i][j]*cmatrix[j][i]*weight[k]/pitimesb;
	}
      }
      dt = obs[k+1] - obs[k];
    }
  }

  for (size_t i = 0; i < p; i++) {
    pi[i] = Bmean[i]/weight_sum;
    if (pi[i] < PMIN) {
      pi[i] = 0.0;
    }
    t[i] = Nmean[i][p]/Zmean[i];
    if (t[i] < PMIN) {
      t[i] = 0.0;
    }
    T[i][i] = -t[i];
    for (size_t j = 0; j < p; j++) {
      if (i != j) {
	T[i][j] = Nmean[i][j]/Zmean[i];
	if (T[i][j] < PMIN) {
	  T[i][j] = 0.0;
	}
	T[i][i] -= T[i][j];
      }
    }
  }
}


static double compute_loglikelihood(const dvec_t& pi, const dmat_t& T,
				    const dvec_t& t, double h) {
  size_t p = pi.size();
  double loglikelihood = 0.0;
  if (obs.size() > 0) {
    dvec_t avector(pi);
    double dt = obs[0];
    for (size_t i = 0; i < obs.size(); i++) {
      if (dt > 0) {
	a_rungekutta(avector, dt, h, T);
      }
      double atimesexit = 0.0;
      for (size_t j = 0; j < p; j++) {
	atimesexit += avector[j] * t[j];
      }
      loglikelihood += weight[i] * log(atimesexit);  
      dt = obs[i+1] - obs[i];
    }
  }
  if (verbosity > 1) {
    std::cout << "Log-likelihood = " << loglikelihood << std::endl;
  }

  return loglikelihood;
}


static void EMiterate(dvec_t& pi, dmat_t& T, dvec_t& t,
		      size_t iters, double epsilon) {
  double h = set_step_length(T);
  double score = compute_loglikelihood(pi, T, t, h);
  if (verbosity == 1) {
    std::cout << "fitting distribution";
  }
  for (size_t k = 1; k <= iters; k++) {
    if (verbosity == 1) {
      if (k % 1000 == 0) {
	std::cout << ':';
      } else if (k % 100 == 0) {
	std::cout << '.';
      }
    }
    EMstep(pi, T, t, h);
    h = set_step_length(T);
    if (k < 6 || (k % 25) == 0 || k == iters) {
      if (verbosity > 2) {
	std::cout << std::endl << "EM(" << k << ")";
	show_pi_T(pi, T, t);
      }
      double new_score = compute_loglikelihood(pi, T, t, h);
      if (k == iters || 1 - new_score/score < epsilon) {
	if (verbosity == 1) {
	  std::cout << ' ' << k << " iterations." << std::endl;
	}
	break;
      }
      score = new_score;
    }
  }
  if (verbosity == 1) {
    std::cout << "Log-likelihood = " << score << std::endl;
  }
}


static double drandom() {
  return 0.9*genrand_real1() + 0.1;
}
 

/* Provides the parameters for a continuous phase-type distribution
   using the EM algorithm with n phases. */
void ph_em_fit(PHParameters& params, const Distribution& dist, size_t n,
	       double discr, double epsilon, size_t iters) {
  dvec_t pi(n);
  dmat_t T(n, n);
  dvec_t t(n);
  double t0 = dist.inv(0);
  obs.clear();
  weight.clear();
  weight_sum = 0.0;
  for (double p = discr; p < 1.0 + 0.5*discr; p += discr) {
    double t = dist.inv(p);
    if (finite(t)) {
      obs.push_back(0.5*(t + t0));
      weight.push_back(discr);
      weight_sum += discr;
      t0 = t;
    }
  }
  std::vector<double> m;
  dist.moments(m, 2);
  double cv2 = m[1]/m[0]/m[0] - 1.0;
  pi[0] = 1.0;
  if (cv2 <= 1.0/n) {
    /* Use Erlang. */
    double lambda = n/m[0];
    for (size_t i = 0; i < n-1; i++) {
      T[i][i] -= T[i][i+1] = lambda;
    }
    T[n-1][n-1] -= t[n-1] = lambda;
  } else {
    /* Use Coxian. */
    for (size_t i = 0; i < n-1; i++) {
      T[i][i] -= t[i] = drandom();
      T[i][i] -= T[i][i+1] = drandom()*n/m[0];
    }
    T[n-1][n-1] -= t[n-1] = drandom()*n/m[0];
  }

  EMiterate(pi, T, t, iters, epsilon);

  params.n = n;
  for (size_t i = 0; i < n; i++) {
    for (size_t j = 0; j < n; j++) {
      if (i != j && T[i][j] > PMIN) {
	params.add_rate(i, j, T[i][j]);
      }
    }
    if (t[i] > PMIN) {
      params.add_rate(i, n, t[i]);
    }
  }
}
