/*
 * Main program.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: main.cc,v 2.3 2005/02/08 11:25:46 lorens Exp $
 */
#include <config.h>
#include "mtbdd.h"
#include "problems.h"
#include "domains.h"
#include "rng.h"
#if HAVE_GETOPT_LONG
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <getopt.h>
#else
#include "getopt.h"
#endif
#include <iostream>
#include <stdexcept>


/* The parse function. */
extern int yyparse();
/* File to parse. */
extern FILE* yyin;

/* Name of current file. */
std::string current_file;
/* Level of warnings. */
int warning_level;
/* Verbosity level. */
int verbosity;

/* Program options. */
static struct option long_options[] = {
  { "tolerance", required_argument, 0, 'E' },
  { "filter", required_argument, 0, 'f' },
  { "discount-factor", required_argument, 0, 'G' },
  { "matching-moments", required_argument, 0, 'm' },
  { "partition", required_argument, 0, 'p' },
  { "sample-paths", required_argument, 0, 'P' },
  { "phases", required_argument, 0, 's' },
  { "seed", required_argument, 0, 'S' },
  { "uniformization", no_argument, 0, 'u' },
  { "verbose", optional_argument, 0, 'v' },
  { "version", no_argument, 0, 'V' },
  { "warnings", optional_argument, 0, 'W' },
  { "help", no_argument, 0, '?' },
  { 0, 0, 0, 0 }
};
static const char OPTION_STRING[] = "E:f:G:m:p:P:s:S:uv::VW::?";


/* Displays help. */
static void display_help() {
  std::cout << "usage: " << PACKAGE << " [options] [file ...]" << std::endl
	    << "options:" << std::endl
	    << "  -E e,  --tolerance=e\t"
	    << "use error tolerance e (default is 0.01)" << std::endl
	    << "  -f f,  --filter=f\t"
	    << "phase filter usage (`n', `h', `t', or `a')" << std::endl
	    << "  -G g,  --discount-factor=g" << std::endl
	    << "\t\t\tuse discount factor g (default is 0.9)" << std::endl
	    << "  -m m,  --matching-moments=m" << std::endl
	    << "\t\t\tmatch the first m moments of general distributions"
	    << std::endl
	    << "  -p p,  --partition=p\t"
	    << "transition relation partition (`f', `e', or `n')" << std::endl
	    << "  -P p,  --sample-paths=p" << std::endl
	    << "\t\t\testimate policy value using p sample paths" << std::endl
	    << "  -s s,  --phases=s\t"
	    << "use s phases to approximate general distributions" << std::endl
	    << "  -S s,  --seed=s\t"
	    << "use s as seed for random number generator" << std::endl
	    << "  -u,    --uniformization" << std::endl
	    << "\t\t\tuse uniformization" << std::endl
	    << "  -v[n], --verbose[=n]\t"
	    << "use verbosity level n;" << std::endl
	    << "\t\t\t  n is a number from 0 (verbose mode off) and up;"
	    << std::endl
	    << "\t\t\t  default level is 1 if optional argument is left out"
	    << std::endl
	    << "  -V,    --version\t"
	    << "display version information and exit" << std::endl
	    << "  -W[n], --warnings[=n]\t"
	    << "determines how warnings are treated;" << std::endl
	    << "\t\t\t  0 supresses warnings; 1 displays warnings;"
	    << std::endl
	    << "  -?,    --help\t\t"
	    << "display this help and exit" << std::endl
	    << "  file ...\t\t"
	    << "files containing domain and problem definitions;" << std::endl
	    << "\t\t\t  if none, definitions are read from standard input"
	    << std::endl
	    << std::endl
	    << "Report bugs to <" PACKAGE_BUGREPORT ">." << std::endl;
}


/* Displays version information. */
static void display_version() {
  std::cout << PACKAGE_STRING << std::endl
	    << "Copyright (C) 2004, 2005 Carnegie Mellon University"
	    << std::endl
	    << PACKAGE_NAME
	    << " comes with NO WARRANTY, to the extent permitted by law."
	    << std::endl
	    << "For information about the terms of redistribution,"
	    << std::endl
	    << "see the file named COPYING in the " PACKAGE_NAME
	    << " distribution." << std::endl
	    << std::endl
	    << "Written by H\345kan L. S. Younes." << std::endl;
}


/* Parses the given file, and returns true on success. */
static bool read_file(const char* name) {
  yyin = fopen(name, "r");
  if (yyin == 0) {
    std::cerr << PACKAGE << ':' << name << ": " << strerror(errno)
	      << std::endl;
    return false;
  } else {
    current_file = name;
    bool success = (yyparse() == 0);
    fclose(yyin);
    return success;
  }
}


/* The main program. */
int main(int argc, char* argv[]) {
  /* Set default verbosity. */
  verbosity = 0;
  /* Set default warning level. */
  warning_level = 0;
  /* Set default tolerance. */
  double epsilon = 0.01;
  /* Set default discount factor. */
  double gamma = 0.9;
  /* Sample size for estimating policy value. */
  int sample_size = 1000;
  /* Transition relation partition. */
  PartitionOption partition = NO_PARTITION;
  /* Phase filtering. */
  FilterOption filter = H_FILTER;
  /* Number of moments to match. */
  size_t moments = 2;
  /* Number of phases to use (if moments is less than 1). */
  size_t phases = 2;

  try {
    /*
     * Get command line options.
     */
    while (1) {
      int option_index = 0;
      int c = getopt_long(argc, argv, OPTION_STRING,
			  long_options, &option_index);
      if (c == -1) {
	break;
      }
      switch (c) {
      case 'E':
	epsilon = atof(optarg);
	if (epsilon <= 0.0) {
	  throw std::invalid_argument("tolerance must be positive");
	}
	break;
      case 'f':
	if (optarg[0] == 'n') {
	  filter = NO_FILTER;
	} else if (optarg[0] == 'h') {
	  filter = H_FILTER;
	} else if (optarg[0] == 't') {
	  filter = T_FILTER;
	} else if (optarg[0] == 'a') {
	  filter = A_FILTER;
	} else {
	  throw std::invalid_argument("filter must be `n', `h', `t', or `a'");
	}
	break;
      case 'G':
	gamma = atof(optarg);
	if (gamma <= 0.0) {
	  throw std::invalid_argument("discount factor must be positive");
	}
	if (gamma >= 1.0) {
	  throw std::invalid_argument("discount factor must be less than 1");
	}
	break;
      case 'p':
	if (partition == UNIFORMIZATION) {
	  std::cerr << PACKAGE
		    << ": cannot use partitioning with uniformization"
		    << std::endl;
	} else {
	  if (optarg[0] == 'f') {
	    partition = FULL_PARTITION;
	  } else if (optarg[0] == 'e') {
	    partition = EVENT_PARTITION;
	  } else if (optarg[0] == 'n') {
	    partition = NO_PARTITION;
	  } else {
	    throw std::invalid_argument("partition must be `f', `e', or `n'");
	  }
	}
	break;
      case 'P':
	sample_size = atoi(optarg);
	if (sample_size < 0) {
	  throw std::invalid_argument("sample size must be positive");
	}
	break;
      case 'm':
	moments = atoi(optarg);
	if (moments > 3) {
	  throw std::invalid_argument("cannot match more than three moments");
	}
	break;
      case 's':
	phases = atoi(optarg);
	if (phases < 1) {
	  throw std::invalid_argument("must use at least one phase");
	}
	moments = 0;
	break;
      case 'S':
	init_genrand(atol(optarg));
	break;
      case 'u':
	partition = UNIFORMIZATION;
	break;
      case 'v':
	verbosity = (optarg != 0) ? atoi(optarg) : 1;
	break;
      case 'V':
	display_version();
	return 0;
      case 'W':
	warning_level = (optarg != 0) ? atoi(optarg) : 1;
	break;
      case '?':
	if (optopt == '?') {
	  display_help();
	  return 0;
	}
      case ':':
      default:
	std::cerr << "Try `" PACKAGE " --help' for more information."
		  << std::endl;
	return 1;
      }
    }

    /*
     * Read input files.
     */
    if (optind < argc) {
      /*
       * Use remaining command line arguments as file names.
       */
      while (optind < argc) {
	if (!read_file(argv[optind++])) {
	  return 1;
	}
      }
    } else {
      /*
       * No remaining command line argument, so read from standard input.
       */
      yyin = stdin;
      if (yyparse() != 0) {
	return 1;
      }
    }

    if (verbosity > 1) {
      /*
       * Display domains and problems.
       */
      std::cerr << "----------------------------------------"<< std::endl
		<< "domains:" << std::endl;
      for (Domain::DomainMap::const_iterator di = Domain::begin();
	   di != Domain::end(); di++) {
	std::cerr << std::endl << *(*di).second << std::endl;
      }
      std::cerr << "----------------------------------------"<< std::endl
		<< "problems:" << std::endl;
      for (Problem::ProblemMap::const_iterator pi = Problem::begin();
	   pi != Problem::end(); pi++) {
	std::cerr << std::endl << *(*pi).second << std::endl;
      }
      std::cerr << "----------------------------------------"<< std::endl;
    }

    /*
     * Solve each problem.
     */
    std::cout.setf(std::ios::unitbuf);
    for (Problem::ProblemMap::const_iterator pi = Problem::begin();
	 pi != Problem::end(); pi++) {
      const Problem& problem = *(*pi).second;
      std::cout << ';' << problem.name() << std::endl;
      DdManager* dd_man = 0;
      mtbdd_solve(dd_man, problem, gamma, epsilon, partition, filter,
		  moments, phases, sample_size);
      if (dd_man != 0) {
	Cudd_DebugCheck(dd_man);
	int unrel = Cudd_CheckZeroRef(dd_man);
	if (unrel != 0) {
	  std::cerr << PACKAGE << ": " << unrel << " unreleased DDs"
		    << std::endl;
	}
	Cudd_Quit(dd_man);
      }
    }
  } catch (const std::exception& e) {
    std::cerr << std::endl << PACKAGE ": " << e.what() << std::endl;
    return 1;
  } catch (...) {
    std::cerr << std::endl << PACKAGE ": fatal error" << std::endl;
    return 1;
  }

  return 0;
}
