/*
 * Problem descriptions.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: problems.cc,v 2.2 2005/02/08 11:41:56 lorens Exp $
 */
#include "problems.h"
#include "domains.h"
#include <typeinfo>


/* ====================================================================== */
/* Problem */

/* Table of defined problems. */
Problem::ProblemMap Problem::problems = Problem::ProblemMap();


/* Returns a const_iterator pointing to the first problem. */
Problem::ProblemMap::const_iterator Problem::begin() {
  return problems.begin();
}


/* Returns a const_iterator pointing beyond the last problem. */
Problem::ProblemMap::const_iterator Problem::end() {
  return problems.end();
}


/* Returns a pointer to the problem with the given name, or 0 if it
   is undefined. */
const Problem* Problem::find(const std::string& name) {
  ProblemMap::const_iterator pi = problems.find(name);
  return (pi != problems.end()) ? (*pi).second : 0;
}


/* Removes all defined problems. */
void Problem::clear() {
  ProblemMap::const_iterator pi = begin();
  while (pi != end()) {
    delete (*pi).second;
    pi = begin();
  }
  problems.clear();
}


/* Constructs a problem. */
Problem::Problem(const std::string& name, const Domain& domain)
  : name_(name), domain_(&domain), terms_(TermTable(domain.terms())),
    goal_(&StateFormula::FALSE), goal_reward_(0), metric_(new Value(0)) {
  RCObject::ref(goal_);
  RCObject::ref(metric_);
  const Problem* p = find(name);
  if (p != 0) {
    delete p;
  }
  problems[name] = this;
}


/* Deletes a problem. */
Problem::~Problem() {
  problems.erase(name());
  for (AtomSet::const_iterator ai = init_atoms_.begin();
       ai != init_atoms_.end(); ai++) {
    RCObject::destructive_deref(*ai);
  }
  for (ValueMap::const_iterator vi = init_values_.begin();
       vi != init_values_.end(); vi++) {
    RCObject::destructive_deref((*vi).first);
  }
  for (EffectList::const_iterator ei = init_effects_.begin();
       ei != init_effects_.end(); ei++) {
    RCObject::destructive_deref(*ei);
  }
  for (RewardRateList::const_iterator ri = reward_rates_.begin();
       ri != reward_rates_.end(); ri++) {
    RCObject::destructive_deref((*ri).first);
  }
  RCObject::destructive_deref(goal_);
  if (goal_reward_ != 0) {
    delete goal_reward_;
  }
  RCObject::destructive_deref(metric_);
  for (EventSet::const_iterator ei = events_.begin();
       ei != events_.end(); ei++) {
    delete *ei;
  }
}


/* Adds an atomic state formula to the initial conditions of this
   problem. */
void Problem::add_init_atom(const Atom& atom) {
  if (init_atoms_.find(&atom) == init_atoms_.end()) {
    init_atoms_.insert(&atom);
    RCObject::ref(&atom);
  }
}


/* Adds a fluent value to the initial conditions of this problem. */
void Problem::add_init_value(const Fluent& fluent, const Rational& value) {
  if (init_values_.find(&fluent) == init_values_.end()) {
    init_values_.insert(std::make_pair(&fluent, value));
    RCObject::ref(&fluent);
  } else {
    init_values_[&fluent] = value;
  }
}


/* Adds an initial effect for this problem. */
void Problem::add_init_effect(const Effect& effect) {
  init_effects_.push_back(&effect);
  RCObject::ref(&effect);
}


/* Adds a reward rate for this problem. */
void Problem::add_reward_rate(const StateFormula& condition,
			      const Rational& reward) {
  reward_rates_.push_back(std::make_pair(&condition, reward));
  RCObject::ref(&condition);
}


/* Sets the goal for this problem. */
void Problem::set_goal(const StateFormula& goal) {
  if (&goal != goal_) {
    RCObject::ref(&goal);
    RCObject::destructive_deref(goal_);
    goal_ = &goal;
  }
}


/* Sets the goal reward for this problem. */
void Problem::set_goal_reward(const Update& goal_reward) {
  if (&goal_reward != goal_reward_) {
    delete goal_reward_;
    goal_reward_ = &goal_reward;
  }
}


/* Sets the metric to maximize for this problem. */
void Problem::set_metric(const Expression& metric, bool negate) {
  const Expression* real_metric;
  if (negate) {
    real_metric = &Subtraction::make(*new Value(0), metric);
  } else {
    real_metric = &metric;
  }
  if (real_metric != metric_) {
    RCObject::ref(real_metric);
    RCObject::destructive_deref(metric_);
    metric_ = real_metric;
  }
}


/* Instantiates this problem. */
void Problem::instantiate() {
  for (RewardRateList::iterator ri = reward_rates_.begin();
       ri != reward_rates_.end(); ri++) {
    const StateFormula& inst_cond =
      (*ri).first->instantiation(SubstitutionMap(), terms(),
				 init_atoms(), init_values());
    RCObject::ref(&inst_cond);
    RCObject::destructive_deref((*ri).first);
    (*ri).first = &inst_cond;
  }
  set_goal(goal().instantiation(SubstitutionMap(), terms(),
				init_atoms(), init_values()));
  if (goal_reward() != 0) {
    set_goal_reward(goal_reward()->instantiation(SubstitutionMap(),
						 init_values()));
  }
  set_metric(metric().instantiation(SubstitutionMap(), init_values()));
  for (EventSchemaMap::const_iterator ei = domain().events().begin();
       ei != domain().events().end(); ei++) {
    (*ei).second->instantiations(events_, terms(),
				 init_atoms(), init_values());
  }
}


/* Tests if the metric is constant. */
bool Problem::constant_metric() const {
  return typeid(metric()) == typeid(Value);
}


/* Output operator for problems. */
std::ostream& operator<<(std::ostream& os, const Problem& p) {
  os << "name: " << p.name();
  os << std::endl << "domain: " << p.domain().name();
  os << std::endl << "objects:" << p.terms();
  os << std::endl << "init:";
  for (AtomSet::const_iterator ai = p.init_atoms().begin();
       ai != p.init_atoms().end(); ai++) {
    os << std::endl << "  " << **ai;
  }
  for (ValueMap::const_iterator vi = p.init_values().begin();
       vi != p.init_values().end(); vi++) {
    os << std::endl << "  (= " << *(*vi).first << ' ' << (*vi).second << ")";
  }
  for (EffectList::const_iterator ei = p.init_effects().begin();
       ei != p.init_effects().end(); ei++) {
    os << std::endl << "  " << **ei;
  }
  os << std::endl << "reward rates:";
  for (RewardRateList::const_iterator ri = p.reward_rates().begin();
       ri != p.reward_rates().end(); ri++) {
    os << std::endl << "  " << *(*ri).first << ' ' << (*ri).second;
  }
  os << std::endl << "goal: " << p.goal();
  if (p.goal_reward() != 0) {
    os << std::endl << "goal reward: " << p.goal_reward()->expression();
  }
  os << std::endl << "metric: " << p.metric();
  os << std::endl << "events:";
  for (EventSet::const_iterator ei = p.events().begin();
       ei != p.events().end(); ei++) {
    os << std::endl << "  " << **ei;
  }
  return os;
}
