/* -*-C++-*- */
/*
 * Problem descriptions.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: problems.h,v 2.1 2005/02/08 11:42:25 lorens Exp $
 */
#ifndef PROBLEMS_H
#define PROBLEMS_H

#include <config.h>
#include "domains.h"
#include "events.h"
#include "effects.h"
#include "formulas.h"
#include "expressions.h"
#include "terms.h"
#include "types.h"
#include <iostream>
#include <map>
#include <string>


/* ====================================================================== */
/* RewardRateList */

/*
 * A list of reward rates.
 */
struct RewardRateList
  : public std::vector<std::pair<const StateFormula*, Rational> > {
};


/* ====================================================================== */
/* Problem */

/*
 * Problem definition.
 */
struct Problem {
  /* Table of problem definitions. */
  struct ProblemMap : public std::map<std::string, const Problem*> {
  };

  /* Returns a const_iterator pointing to the first problem. */
  static ProblemMap::const_iterator begin();

  /* Returns a const_iterator pointing beyond the last problem. */
  static ProblemMap::const_iterator end();

  /* Returns a pointer to the problem with the given name, or 0 if it
     is undefined. */
  static const Problem* find(const std::string& name);

  /* Removes all defined problems. */
  static void clear();

  /* Constructs a problem. */
  Problem(const std::string& name, const Domain& domain);

  /* Deletes a problem. */
  ~Problem();

  /* Returns the name of this problem. */
  const std::string& name() const { return name_; }

  /* Returns the domain of this problem. */
  const Domain& domain() const { return *domain_; }

  /* Returns the term table of this problem. */
  TermTable& terms() { return terms_; }

  /* Returns the term table of this problem. */
  const TermTable& terms() const { return terms_; }

  /* Adds an atomic state formula to the initial conditions of this
     problem. */
  void add_init_atom(const Atom& atom);

  /* Adds a fluent value to the initial conditions of this problem. */
  void add_init_value(const Fluent& fluent, const Rational& value);

  /* Adds an initial effect for this problem. */
  void add_init_effect(const Effect& effect);

  /* Adds a reward rate for this problem. */
  void add_reward_rate(const StateFormula& condition, const Rational& reward);

  /* Sets the goal for this problem. */
  void set_goal(const StateFormula& goal);

  /* Sets the goal reward for this problem. */
  void set_goal_reward(const Update& goal_reward);

  /* Sets the metric to maximize for this problem. */
  void set_metric(const Expression& metric, bool negate = false);

  /* Instantiates this problem. */
  void instantiate();

  /* Returns the initial atoms of this problem. */
  const AtomSet& init_atoms() const { return init_atoms_; }

  /* Returns the initial values of this problem. */
  const ValueMap& init_values() const { return init_values_; }

  /* Returns the initial effects of this problem. */
  const EffectList& init_effects() const { return init_effects_; }

  /* Returns the reward rates of this problem. */
  const RewardRateList& reward_rates() const { return reward_rates_; }

  /* Returns the goal of this problem. */
  const StateFormula& goal() const { return *goal_; }

  /* Returns a pointer to the goal reward for this problem, or 0 if no
     explicit reward is associated with goal states. */
  const Update* goal_reward() const { return goal_reward_; }

  /* Returns the metric to maximize for this problem. */
  const Expression& metric() const { return *metric_; }

  /* Tests if the metric is constant. */
  bool constant_metric() const;

  /* Returns a set of instantiated events. */
  const EventSet& events() const { return events_; }

private:
  /* Table of defined problems. */
  static ProblemMap problems;

  /* Name of problem. */
  std::string name_;
  /* Problem domain. */
  const Domain* domain_;
  /* Problem terms. */
  TermTable terms_;
  /* Initial atoms. */
  AtomSet init_atoms_;
  /* Initial fluent values. */
  ValueMap init_values_;
  /* Initial effects. */
  EffectList init_effects_;
  /* Reward rates. */
  RewardRateList reward_rates_;
  /* Goal; FALSE if not a goal-directed planning problem. */
  const StateFormula* goal_;
  /* Goal reward expression. */
  const Update* goal_reward_;
  /* Metric to maximize. */
  const Expression* metric_;
  /* Instantiated events. */
  EventSet events_;
};

/* Output operator for problems. */
std::ostream& operator<<(std::ostream& os, const Problem& p);


#endif /* PROBLEMS_H */
